{-# LANGUAGE BangPatterns #-}
module Mandel where
import Complex -- 1.3
import PortablePixmap
import Control.Parallel
import Control.Parallel.Strategies (using)

default ()


mandel::(Num a) => a -> [a]
mandel c = infiniteMandel
           where
		infiniteMandel = c : (map (\z -> z*z +c) infiniteMandel)


whenDiverge::  Int -> Double -> Complex Double -> Int
whenDiverge limit radius c
  = walkIt (take limit (mandel c))
  where
     walkIt []     = 0					 -- Converged
     walkIt (x:xs) | diverge x radius  = 0		 -- Diverged
	           | otherwise		 = 1 + walkIt xs -- Keep walking

-- VERY IMPORTANT FUNCTION: sits in inner loop

diverge::Complex Double -> Double -> Bool
diverge cmplx radius =  magnitude cmplx > radius

parallelMandel:: [[Complex Double]] -> Int -> Double -> [Int]
parallelMandel mat limit radius
   = concat $ 

                 parBuffer 70
                             [ let l = map (whenDiverge limit radius) xs
                               in seqList l `pseq` l
                             | xs <- mat ]


parBuffer :: Int -> [a] -> [a]
parBuffer n xs = return xs (start n xs)
  where
    return (x:xs) (y:ys) = y `par` (x : return xs ys) 
    return xs [] = xs

    start !n [] = []
    start 0 ys = ys
    start !n (y:ys) = y `par` start (n-1) ys



lazyParList :: Int -> [a] -> [a]
lazyParList !n xs = go xs (parListN n xs)
   where 
         go []     _ys    = []
         go (x:xs) []     = x : xs
         go (x:xs) (y:ys) = y `par` (x : go xs ys)

lazyParList1 :: Int -> [a] -> [a]
lazyParList1 !n xs = go xs (parListN1 n xs [])
   where 
         go []     _ys    = []
         go (x:xs) []     = x : xs
         go (x:xs) (y:ys) = y `par` (x : go xs ys)


parList :: [a] -> ()
parList [] = ()
parList (x:xs) = x `par` parList xs


parListN :: Int -> [a] -> [a]
parListN 0  xs     = xs 
parListN !n []     = []
parListN !n (x:xs) = x `par` parListN (n-1) xs


parListN1 :: Int -> [a] -> [a] -> [a]
parListN1 0  xs     ys = parList ys `pseq` xs
parListN1 !n []     ys = parList ys `pseq` []
parListN1 !n (x:xs) ys = parListN1 (n-1) xs (x:ys)

seqList :: [a] -> ()
seqList [] = ()
seqList (x:xs) = x `pseq` seqList xs



mandelset::Double -> 			-- Minimum X viewport
	   Double -> 			-- Minimum Y viewport
	   Double -> 			-- Maximum X viewport
	   Double ->			-- maximum Y viewport
	   Integer -> 			-- Window width
	   Integer -> 			-- Window height
	   Int -> 			-- Window depth
	   PixMap			-- result pixmap
mandelset x y x' y' screenX screenY lIMIT
   = createPixmap screenX screenY lIMIT (map prettyRGB result)
   where

      windowToViewport s t
           = ((x + (((coerce s) * (x' - x)) / (fromInteger screenX))) :+
	      (y + (((coerce t) * (y' - y)) / (fromInteger screenY))))

      coerce::Integer -> Double
      coerce  s   = encodeFloat (toInteger s) 0

      result = parallelMandel
		  [[windowToViewport s t | s<-[1..screenX]]
                   | t <- [1..screenY]]
 		  lIMIT
		  ((max (x'-x) (y'-y)) / 2.0)

      prettyRGB::Int -> (Int,Int,Int)
      prettyRGB s = let t = (lIMIT - s) in (s,t,t)

