{-# OPTIONS -fglasgow-exts #-}

module Main(main) where

import Data.List(transpose)
import System(getArgs)
import GHC.Base(par#, parGlobal#, Int#)
import GHC.Conc(par,pseq)
import Control.Monad
import System.Time
import Text.Printf


main = do args <- getArgs
          let 
            n = read (args!!0) :: Int -- size of the interval
            c = read (args!!1) :: Int -- chunk size
          t1 <- getClockTime
          let res = sumEuler 1 n c
          putStrLn ("sumEuler over an interval from 1 to " ++ (show n) ++ " with chunk size " ++ (show c) ++ " = " ++ (show res))
          t2 <- getClockTime
          putStrLn $ printf "time taken: %.2fs" $ secDiff t1 t2
              
		
		
          
-- sumEuler over an interval from m to n (with chunk size c)
sumEuler ::  Int -> Int -> Int -> Int
sumEuler m n c | m>n = 0 
	       | otherwise = let 
                              this = sum (map euler [m..m+c-1])
                              rest = sumEuler (min (m+c) (n+1)) n c
                             in
                              this `par` (rest `pseq` this+rest)

sumEuler_seq :: Int -> Int
sumEuler_seq = sum . map euler . enumFromTo 1

euler :: Int -> Int
euler n = let
            relPrimes = let
                          numbers = [1..(n-1)]
                        in
                          {- numbers `par` -} (filter (relprime n) numbers)
          in
            {- (spine relPrimes) `par`-} (length relPrimes)

-- aux fcts
hcf :: Int -> Int -> Int
hcf x 0 = x
hcf x y = hcf y (rem x y)

relprime :: Int -> Int -> Bool
relprime x y = hcf x y == 1

-- strategic functions (could uses Strategies module instead)
parList :: [Int] -> ()
parList = foldr par ()

spine :: [Int] -> ()
spine []     = ()
spine (_:xs) = spine xs

chunk :: Int -> [a] -> [[a]]
chunk _n [] = []
chunk n  xs = ys:chunk n zs where (ys,zs) = splitAt n xs

slice :: Int ->  [a] -> [[a]]
slice n = transpose . chunk n

-- func to calc time taken between t0 and t1 in sec
secDiff::ClockTime -> ClockTime -> Float
secDiff (TOD secs1 psecs1) (TOD secs2 psecs2) = fromInteger (psecs2 - psecs1) / 1e12 + fromInteger (secs2 - secs1)
