{-# OPTIONS -cpp -fglasgow-exts #-}

module Main(main) where

import System(getArgs)
import Debug.Trace(trace)
import Control.Monad
import GHC.Base(par#, parGlobal#, Int#)
import GHC.Conc(par,pseq)
import System.Time
import Text.Printf

main = do args <- getArgs
          unless (length args == 3) $
            error "Usage: worpitzky <x> <n> <t>  ... testing Worpitzky identity for x^n, using t as a threshold"
          let 
            x :: Integer
            x = read (args!!0)  -- input for worpitzky
            n :: Integer
            n = read (args!!1)  -- input for worpitzky
            t :: Integer
            t = read (args!!2)  -- threshold
          t1 <- getClockTime
          let res = worpitzky x n t
          putStrLn ("Testing Worpitzky's identity for x^n ..." ++ (show x) ++ " " ++ (show n) ++ " " ++ (show t) ++ " = " ++ (show res))
          t2 <- getClockTime
          putStrLn $ printf "time taken: %.2fs" $ secDiff t1 t2

worpitzky ::  Integer -> Integer -> Integer -> Bool
worpitzky x n t = x^n == sum [ (par_euler n k t) * (bin (x+k) n) | k <- [0,1..n] ]

-- euler numbers (Concrete Maths, p254)
euler :: Integer -> Integer -> Integer
euler 0 k = if k==0 then 1 else 0
euler n k = (k+1) * (euler (n-1) k) + (n-k)*(euler (n-1) (k-1))

-- binomials 
bin :: Integer -> Integer -> Integer
bin n k = fact n `div` ( (fact k) * (fact (n-k)) )

fact :: Integer -> Integer
fact 0 = 1
fact n = product [1..n]

par_euler :: Integer -> Integer -> Integer -> Integer
par_euler 0 k t = if k==0 then 1 else 0
par_euler n k t | n<t = euler n k
par_euler n k t = e2 `par` (e1 `pseq` ((k+1) * e1 + (n-k)*e2))
                  where e1 = par_euler (n-1) k t
                        e2 = par_euler (n-1) (k-1) t

stir1 :: Integer -> Integer -> Integer
stir1 _ 0 = 1
stir1 0 _ = 1
stir1 n k = (n-1)*(stir1 (n-1) k)+(stir1 (n-1) (k-1))

par_stir1 :: Integer -> Integer -> Integer -> Integer
par_stir1 n k t | n<t || k<t = stir1 n k
                | otherwise  = s1 `par` (s2 `par` (n-1)*s1+s2)
                	       where s1 = stir1 (n-1) k
                	             s2 = stir1 (n-1) (k-1)

stir2 :: Integer -> Integer -> Integer
stir2 _ 1 = 1
stir2 0 _ = 1
stir2 n k = (stir2 (n-1) (k-1)) + k*(stir2 (n-1) k)

par_stir2 :: Integer -> Integer -> Integer -> Integer
par_stir2 n k t | n<t || k<t = stir2 n k
                | otherwise  = s1 `par` (s2 `pseq` s1+k*s2)
                	       where s1 = par_stir2 (n-1) (k-1) t
                	             s2 = k*(par_stir2 (n-1) k t)

-- func to calc time taken between t0 and t1 in sec
secDiff::ClockTime -> ClockTime -> Float
secDiff (TOD secs1 psecs1) (TOD secs2 psecs2) = fromInteger (psecs2 - psecs1) / 1e12 + fromInteger (secs2 - secs1)
